#
# Copyright 2011 Eigenlabs Ltd.  http://www.eigenlabs.com
#
# This file is part of EigenD.
#
# EigenD is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EigenD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EigenD.  If not, see <http://www.gnu.org/licenses/>.
#

import piw
from pi import agent, domain, bundles,atom
#TODO: why is the version not taken from the concrete classes, this line should not be necessary...
from plg_midiilluminator import tau_midi_light_controller_version as version

#
# midiilluminator_plg_native is generated by the build system through the midiilluminator_plg.pip file.
#
from midiilluminator_plg_native import midiilluminator

#
# The midiilluminator lights agent
#
class IlluminatorAgent(agent.Agent):
    def __init__(self, address, ordinal, num_keys, name):
        agent.Agent.__init__(self, signature=version, names=name, ordinal=ordinal)

        self.__num_layouts = 20
        self.__num_keys = num_keys

        self.domain = piw.clockdomain_ctl()
        self.domain.set_source(piw.makestring('*',0))
        
        self[2] = atom.Atom(names="outputs")
        self[2][1] = bundles.Output(1,False,names='light output')
        
        self.output = bundles.Splitter(self.domain,*self[2].values())

        self.midiilluminator = midiilluminator(self.domain, self.output.cookie(), num_keys)
        
        self.data_input = bundles.VectorInput(self.midiilluminator.cookie(), self.domain, signals=(1,))
        self[1] = atom.Atom(names="inputs")
        self[1][1] = atom.Atom(domain=domain.BoundedFloat(-1,1), policy=self.data_input.vector_policy(1,False), names='midi input')
        
        self[3] = atom.Atom(domain=domain.BoundedInt(-120,120), init=0, policy=atom.default_policy(self.__set_transpose), names='transpose')
        
        self[4] = atom.Atom(domain=domain.BoundedInt(1,self.__num_layouts), init=1, policy=atom.default_policy(self.__set_layout), names='current layout')
        
        self.__LIGHT_NAME = ['green', 'orange', 'red']
        
        for color in range(0,3): # green, orange, red
            self[5+color] = atom.Atom(names=self.__LIGHT_NAME[color] + " lights")
            self[5+color][1] = atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   # scoping in Python is lexical, so we have to create an explicit scope with a second lambda
                                   # to have the expected variable values in the closure
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_direct_keys(color, state))(color)),
                                   names='enable direct keys')
            self[5+color][2] = atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_direct_notes(color, state))(color)),
                                   names='enable direct notes')
            self[5+color][3] = atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_midi(color, state))(color)),
                                   names='enable midi')
        self[8] = atom.Atom(names="keys")
        for key in range(1,num_keys + 1):
            self[8][key] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                   init=0,
                                   policy=atom.default_policy((lambda key: lambda val: self.__set_key(key,val))(key)),
                                   names='key ' + `key`)
        self[9] = atom.Atom(names="notes")
        notenum = 0
        for octave in range(0,11):
            for note in ['c', 'cis', 'd', 'dis', 'e', 'f', 'fis', 'g', 'gis', 'a', 'ais', 'b']:
                if notenum > 127:
                    break
                self[9][notenum + 1] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                           init=0,
                                           policy=atom.default_policy((lambda notenum: lambda val: self.__set_note(notenum,val))(notenum)),
                                           names=`notenum + 1` + ': ' + note + `octave`)
                notenum = notenum + 1
        self[10] = atom.Atom(names = "midi channels")    
        for channel in range(1,17):
            self[10][channel] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                          init=0,
                                          policy=atom.default_policy((lambda channel: lambda color: self.__set_channel_color(channel, color))(channel)),
                                          #policy=atom.default_policy(self.__dummy),
                                          names='color for channel ' + `channel`)
            
        self[11] = atom.Atom(names="key to note maps")
        self[11][1] = atom.Atom(names="current layout")
        for key in range(1,num_keys + 1):
                    self[11][1][key] = atom.Atom(domain=domain.BoundedInt(0,128), #0: no note, 1-128: note number
                                               init=self._default_note(key),
                                               policy=atom.default_policy((lambda key: lambda notenum: self.__set_note_for_key(1, key, notenum))(key)),
                                               names='key ' + `key`)
        for layout in range(2,self.__num_layouts + 2):
            self[11][layout] = atom.Atom(names="layout " + `layout - 1`)
            for key in range(1,num_keys + 1):
                    self[11][layout][key] = atom.Atom(domain=domain.BoundedInt(0,128), #0: no note, 1-128: note number
                                               init=self._default_note(key),
                                               policy=atom.default_policy((lambda layout, key: lambda notenum: self.__set_note_for_key(layout, key, notenum))(layout, key)),
                                               names='key ' + `key`)
        self[12] = atom.Atom(domain=domain.Bool(),
                             init=True,
                             policy=atom.default_policy(self.__set_enable),
                             names="enable")
        self[13] = atom.Atom(names="priorities")
        self[13][1] = atom.Atom(domain=domain.BoundedInt(1,4),
                                init=3,
                                policy=atom.default_policy(self.__set_direct_keys_prio),
                                names='direct keys prio')
        self[13][2] = atom.Atom(domain=domain.BoundedInt(1,4),
                                init=2,
                                policy=atom.default_policy(self.__set_direct_notes_prio),
                                names='direct notes prio')
        self[13][3] = atom.Atom(domain=domain.BoundedInt(1,4),
                                init=1,
                                policy=atom.default_policy(self.__set_midi_prio),
                                names='midi prio')
        
    def agent_postload(self):
        if self[12].get_value() == True:
            self.__set_enable(False)
            self.__set_enable(True)
                             
    def _default_note(self, key):
        pass
  
    def __set_transpose(self,t):
        self[3].set_value(t)
        self.midiilluminator.set_transpose(t)
        return True
        
    def __set_layout(self, new_layout):
        old_layout = self[4].get_value()
        if new_layout <> old_layout:
            self[4].set_value(new_layout)
            for key in range(1,self.__num_keys):
                self[11][1][key].set_value(self[11][new_layout + 1][key].get_value())
            for key in range(1,self.__num_keys):
                self.midiilluminator.set_note_for_key(key, self[11][1][key].get_value())
        return True
    
    def __set_enable_direct_keys(self, color, enable):
        self[5+color][1].set_value(enable)
        self.midiilluminator.set_enable_direct_keys_for_color(color + 1, enable)
        return True
    
    def __set_enable_direct_notes(self, color, enable):
        self[5+color][2].set_value(enable)
        self.midiilluminator.set_enable_direct_notes_for_color(color + 1, enable)
        return True
    
    def __set_enable_midi(self, color, enable):
        self[5+color][3].set_value(enable)
        self.midiilluminator.set_enable_midi_for_color(color + 1, enable)
        return True
    
    def __set_key(self, key, val):
        self[8][key].set_value(val)
        self.midiilluminator.set_key(key,val)
        return True
    
    def __set_note(self, note, val):
        self[9][note + 1].set_value(val)
        self.midiilluminator.set_note(note,val)
        return True
    
    def __set_channel_color(self, channel, color):
        self[10][channel].set_value(color)
        self.midiilluminator.set_channel_color(channel, color)
        return True
    
    def __set_note_for_key(self, layout, key, note):
        self[11][layout][key].set_value(note)
        active_layout = self[4].get_value() + 1
        if layout == active_layout:
            self[11][1][key].set_value(note)
        if layout == 1:
            self[11][active_layout][key].set_value(note)
        if (layout == active_layout ) or (layout == 1):
            self.midiilluminator.set_note_for_key(key, note)
        return True
    
    def __set_enable(self, state):
        self[12].set_value(state)
        self.midiilluminator.set_enable(state)
        return True
    
    def __set_direct_keys_prio(self, prio):
        self[13][1].set_value(prio)
        self.midiilluminator.set_direct_keys_prio(prio)
        return True
    
    def __set_direct_notes_prio(self, prio):
        self[13][2].set_value(prio)
        self.midiilluminator.set_direct_notes_prio(prio)
        return True
    
    def __set_midi_prio(self, prio):
        self[13][3].set_value(prio)
        self.midiilluminator.set_midi_prio(prio)
        return True
