
/*
 Copyright 2009 Eigenlabs Ltd.  http://www.eigenlabs.com

 This file is part of EigenD.

 EigenD is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 EigenD is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with EigenD.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <piw/piw_bundle.h>
#include <piw/piw_data.h>
#include <piw/piw_clock.h>
#include <piw/piw_tsd.h>
#include <picross/pic_ilist.h>

#include "simpleagent.h"

//simpleagent_impl_t method definitions
simpleagent_impl_t::simpleagent_impl_t(piw::clockdomain_ctl_t *cd, const piw::cookie_t &c, const char* name)
	: root_t(0), up_(0)
{
	connect(c);
	cd->sink(this,name);
	tick_enable(true);
	set_clock(this);
}

simpleagent_impl_t::~simpleagent_impl_t()
{
	tracked_invalidate();
	invalidate();
}

void simpleagent_impl_t::clocksink_ticked(unsigned long long f, unsigned long long t)
{
	simpleagent_wire_t *w;

	for(w=tickers_.head(); w!=0; w=tickers_.next(w))
	{
		w->ticked(f,t);
	}
}

void simpleagent_impl_t::add_ticker(simpleagent_wire_t *w)
{
	if(!tickers_.head())
	{
		tick_suppress(false);
	}

	tickers_.append(w);
}

void simpleagent_impl_t::del_ticker(simpleagent_wire_t *w)
{
	tickers_.remove(w);

	if(!tickers_.head())
	{
		tick_suppress(true);
	}

}

void simpleagent_impl_t::invalidate()
{
	pic::lckmap_t<piw::data_t,simpleagent_wire_t *>::lcktype::iterator ci;
	tick_disable();

	while((ci=children_.begin())!=children_.end())
	{
		delete ci->second;
	}
}

piw::wire_t *simpleagent_impl_t::root_wire(const piw::event_data_source_t &es)
{
   pic::lckmap_t<piw::data_t,simpleagent_wire_t *>::lcktype::iterator ci;

	if((ci=children_.find(es.path()))!=children_.end())
	{
		delete ci->second;
	}

	simpleagent_wire_t* new_wire = create_wire(es);
	if(new_wire)
		current_wire_.assign(new_wire);
	return new_wire;
}

void simpleagent_impl_t::root_closed()
{
	current_wire_.clear();
	invalidate();
}

void simpleagent_impl_t::root_opened()
{
	root_clock();
	root_latency();
}

void simpleagent_impl_t::root_clock()
{
	if(up_)
	{
		remove_upstream(up_);
		up_=0;
	}

	up_=get_clock();

	if(up_)
	{
		add_upstream(up_);
	}
}

void simpleagent_impl_t::root_latency()
{
	set_latency(get_latency());
}


/*
 * Static method to be called by the fast thread.
 */
int __wire_invalidator(void *w_, void *_)
{
	simpleagent_wire_t *w = (simpleagent_wire_t *)w_;
    if(w->root_)
    {
        w->root_->del_ticker(w);
    }
    return 0;
};

// simpleagent_wire_t method definitions
simpleagent_wire_t::simpleagent_wire_t(simpleagent_impl_t *p, const piw::event_data_source_t& es, unsigned int num_out_signals)
	: piw::event_data_source_real_t(es.path()), root_(p), last_from_(0), num_out_signals_(num_out_signals)
{
	root_->children_.insert(std::make_pair(path(),this));
	root_->connect_wire(this, source());
	subscribe_and_ping(es);
}


simpleagent_wire_t::~simpleagent_wire_t()
{
	//tracked_invalidate();
	invalidate();
}

void simpleagent_wire_t::wire_closed()
{
	delete this;
}

void simpleagent_wire_t::invalidate()
{
	source_shutdown();
	unsubscribe();
    //piw::wire_t::disconnect();
    //wire_ctl_t::disconnect();
	piw::tsd_fastcall(__wire_invalidator, this, 0);
	if(root_)
	{
		root_->children_.erase(path());
		root_ = 0;
	}
}

void simpleagent_wire_t::event_start(unsigned seq, const piw::data_nb_t &id, const piw::xevent_data_buffer_t &b)
{
	output_ = piw::xevent_data_buffer_t(num_out_signals_,PIW_DATAQUEUE_SIZE_NORM);
	input_ = b.iterator();

	unsigned long long t = id.time();
	last_from_ = t;

	pic::logmsg() << "event_start: seq=" << seq << " time: " << t;

	source_start(seq,id,output_);
	root_->add_ticker(this);
}

bool simpleagent_wire_t::event_end(unsigned long long t)
{
	// Process the data that was still present at event end
	ticked(last_from_,t);

	// Remove ourselves from the wires that are processed during a clock tick
	root_->del_ticker(this);

	pic::logmsg() << "event_end: time: " << t;

	// End our own event on our own data source for downstream agents
	return source_end(t);
}

void simpleagent_wire_t::event_buffer_reset(unsigned s, unsigned long long t, const piw::dataqueue_t &o, const piw::dataqueue_t &n)
{
	input_->set_signal(s,n);
	input_->reset(s,t);
	ticked(last_from_,t);
}

//virtual void process(unsigned, const piw::data_nb_t &, unsigned long long) = 0;

void simpleagent_wire_t::ticked(unsigned long long f, unsigned long long t)
{
	last_from_ = t;

	piw::data_nb_t d;
	unsigned s;

	while(input_->next(num_out_signals_,s,d,t))
	{
		(this)->process(s,d,t);
	}
}

void simpleagent_wire_t::source_ended(unsigned seq)
{
	event_ended(seq);
}


